from matrix_mpi import MatrixMPI as Matrix
from vector_mpi import VectorMPI as Vector

# from matrix import Matrix
# from vector import Vector


def cg(A: Matrix, x0: Vector, b: Vector, tolerance: float = 1e-3, max_iterations: int = 1_000):
    """
    Solves a system of linear equations of the form Ax = b numerically.
    
    :param A: The transformation matrix A
    :param x0: A vector to start the algorithm with
    :param b: The solution vector of the system of linear equations, the right hand side
    :param tolerance: The tolerance at which to stop the algorithm, default is 0.001
    :param max_iterations: Maximum number of iterations, default is 1000
    """
    iterations = 0

    x = x0
    r = b - A * x
    d = r

    while r.norm() >= tolerance and iterations < max_iterations:
        z = A * d

        alpha = (r.T() * d) / (d.T() * z)
        x = x + alpha * d
        r = r - alpha * z

        beta = -(r.T() * z) / (d.T() * z)
        d = r + beta * d

        iterations = iterations + 1
    return x
